## Sid Meier's Civilization 4
## Copyright Firaxis Games 2005
##
## GeneralMoraleUtils by TheLopez
## Derived from StackAid by Zuul
##

from CvPythonExtensions import *
import CvUtil
import Popup as PyPopup
import math
import PyHelpers
import TheLopezUtils
import CvConfigParser

import pickle

################# SD-UTILITY-PACK ###################
import SdToolKit
sdEcho         = SdToolKit.sdEcho
sdModInit      = SdToolKit.sdModInit
sdModLoad      = SdToolKit.sdModLoad
sdModSave      = SdToolKit.sdModSave
sdEntityInit   = SdToolKit.sdEntityInit
sdEntityExists = SdToolKit.sdEntityExists
sdGetVal       = SdToolKit.sdGetVal
sdSetVal       = SdToolKit.sdSetVal


# globals
gc = CyGlobalContext()

LAST_X = "LAST_X"
LAST_Y = "LAST_Y"

PyPlayer = PyHelpers.PyPlayer

# Change the variable to either enable or disable the exuding of morale from 
# Great Generals who have joined cities.
# Default value is true
g_bSettledGreatGeneralsExudeMorale = true

class GeneralMoraleUtils:		
	
	def __init__(self):
		# Initialize the class
		
		global g_bSettledGreatGeneralsExudeMorale
		
		config = CvConfigParser.CvConfigParser("Great General Amped Config.ini")

		if(config != None):
			g_bSettledGreatGeneralsExudeMorale = config.getboolean("Great General Amped", "Settled Great Generals Exude Morale", True)


	# Updates the target plot with the correct
	def updateGeneralPlot(self, iPlayerID, objTargetPlot):

		# Get the player
		objPlayer = gc.getPlayer(iPlayerID)

		# Try to get the closest Great General to the target plot that belongs 
		# to the player
		objGeneralPlot = self.getClosestGeneralPlot(objPlayer, objTargetPlot)

		# Return immediately if no Great General plot for the player was found
		if(objGeneralPlot == None):
			return

		# Get the distance between the target plot and the found Great General
		iGeneralDistance = TheLopezUtils.getPlotDistance(objGeneralPlot, objTargetPlot)
		
		# Remove all the morale promotions from the units found in the target 
		# plot
		self.removeAllMoralePromotions(objPlayer, objTargetPlot)

		# Don't promote the units if they are out of the range of any generals
		if(iGeneralDistance >= 3):
			return
		
		# Set the morale promotions for the units found in the target plot
		self.setMoralePromotions(objPlayer, objTargetPlot, iGeneralDistance)


	# Updates all of the units in the plot passed in
	def updatePlot(self, iPlayerID, objPlot):

		# Get the player
		objPlayer = gc.getPlayer(iPlayerID)

		# Return immediately if an invalid player ID was passed in
		if(objPlayer == None):
			return
		
		# Get the closest Great General to the plot belonging to the player
		objGeneralPlot = self.getClosestGeneralPlot(objPlayer, objPlot)

		# Return immediately if no Great General plot for the player was found
		if(objGeneralPlot == None):
			return

		# Get the distance between the plot and the Great General found
		iGeneralDistance = TheLopezUtils.getPlotDistance(objGeneralPlot, objPlot)

		# Remove all unit morale promotions 
		self.removeAllMoralePromotions(objPlayer, objPlot)
		
		# Don't promote them if they are out of the range of any generals 		
		if(iGeneralDistance >= 3):
			return
		
		# Set the morale promotions for the units in the plot
		self.setMoralePromotions(objPlayer, objPlot, iGeneralDistance)


	# Saves the generals location for later use
	def saveGeneralLocation(self, objUnit):
	
		# Return immediately if the unit passed in isn't valid
		if(objUnit == None):
			return
	
		# Return immediately if the unit passed in isn't a general
		if(objUnit.getUnitType() != gc.getInfoTypeForString("UNIT_GENERAL")):
			return

		# If the Great General location information hasn't been previously 
		# logged, then set up the information.
		if(sdEntityExists("Great General Amped",objUnit.getID()) == False):
			self.setupScript(objUnit)
			return
		
		# Save the Great General's location information
		sdSetVal("Great General Amped", objUnit.getID(), LAST_X, objUnit.getX())
		sdSetVal("Great General Amped", objUnit.getID(), LAST_Y, objUnit.getY())
		
				
	# Sets up the Great General entity location information
	def setupScript(self, objGeneral):
		scriptDict = { 	LAST_X : objGeneral.plot().getX(),
						LAST_Y : objGeneral.plot().getY() }
		sdEntityInit("Great General Amped", objGeneral.getID(), scriptDict)
				
		
	# Remove all of the morale promotions from all units found in the
	# objPlot.
	def removeAllMoralePromotions(self, player, objPlot):
		
		# Get the list of units found in the plot
		unitList = TheLopezUtils.getPlayerPlotUnitList(objPlot, player.getID())
		
		iPromotionMorale1 = gc.getInfoTypeForString("PROMOTION_MORALE1")
		iPromotionMorale2 = gc.getInfoTypeForString("PROMOTION_MORALE2")
		iPromotionMorale3 = gc.getInfoTypeForString("PROMOTION_MORALE3")

		promotionList = [iPromotionMorale3,iPromotionMorale2,iPromotionMorale1]

		# Go through the list of units found in the objPlot
		for objUnit in unitList:

			# Go through each of the morale promotions and remove
			# them from the units found in objPlot
			for iPromotionInfoID in promotionList:
				objUnit.setHasPromotion(iPromotionInfoID, false)
				

	# Set the morale promotions for the units found in the objPlot using
	# the iDistance from the nearest general
	def setMoralePromotions(self, player, objPlot, iDistance):
		
		# Get the list of units found in the plot
		unitList = TheLopezUtils.getPlayerPlotUnitList(objPlot, player.getID())

		iPromotionMorale1 = gc.getInfoTypeForString("PROMOTION_MORALE1")
		iPromotionMorale2 = gc.getInfoTypeForString("PROMOTION_MORALE2")
		iPromotionMorale3 = gc.getInfoTypeForString("PROMOTION_MORALE3")

		promotionList = [iPromotionMorale3,iPromotionMorale2,iPromotionMorale1]
 
		# Return immediately if getPlayerPlotUnitList returned a null
		# value for the unitList 
		if(unitList == None):
			return

		# Return immediately if the getPlayerPlotUnitList returns an
		# empty list.
		if(len(unitList) <= 0):
			return

		# Go through the list of units found in the objPlot
		for objUnit in unitList:
			# Here's the tricky concept, only assign the promotions 
			# in relation to the distance from the nearest general
			# so, if the distance from the general is two plots then
			# only set the promotions starting from the 3rd value
			# in the m_promotionList. Currently the promotion list 
			# contains three values so only the third promotion would
			# be given. 
			# Distance from General			Promotions Given
			# 	0					PROMOTION_MORALE1
			#						PROMOTION_MORALE2
			#						PROMOTION_MORALE3
			#
			# 	1					PROMOTION_MORALE1
			#						PROMOTION_MORALE2
			#
			# 	2					PROMOTION_MORALE1
			for j in range(iDistance, len(promotionList)):
				if(gc.getUnitInfo(objUnit.getUnitType()).getDomainType() == DomainTypes.DOMAIN_LAND):
					objUnit.setHasPromotion(promotionList[j], true)	
					
					
	# Returns the closest plot containing a general to the plot that belongs to 
	# the player
	def getClosestGeneralPlot(self, objPlayer, objPlot):
		' CvUnit - the closest General to the plot'
		
		objGeneralPlot = None
		
		# Make the starting distance bigger than any map should ever get.
		# If maps ARE wider than this then we will all be in trouble and
		# need super computers to play the game.
		iDistance = 999999
				
		# Get the player's list of generals
		generalList = PyPlayer(objPlayer.getID()).getUnitsOfType(gc.getInfoTypeForString("UNIT_GENERAL"))

		for objGeneral in generalList:
			iTempDistance = TheLopezUtils.getPlotDistance(objPlot, objGeneral.plot())

			# If the general is closer than the current distance then set the 
			# plot their at as the closest general plot and save their 
			# distance as the current distance
			if(iTempDistance < iDistance):
				iDistance = iTempDistance
				objGeneralPlot = objGeneral.plot()

		
		# Search the player's cities if the player has configured for Great
		# Generals that have joined cities to exude morale
		if(g_bSettledGreatGeneralsExudeMorale):
			cityList = TheLopezUtils.getPlayerCitiesContainingSpecialist(gc.getInfoTypeForString("SPECIALIST_GREAT_GENERAL"),objPlayer.getID())

			for objCity in cityList:
				iTempDistance = TheLopezUtils.getPlotDistance(objPlot, objCity.plot())

				# If the general is closer than the current distance then set 
				# the plot their at as the closest general plot and save their 
				# distance as the current distance
				if(iTempDistance < iDistance):
					iDistance = iTempDistance
					objGeneralPlot = objCity.plot()

		return objGeneralPlot


	# Handles the cases where a unit has moved from one plot to another
	def handleUnitMove(self, objUnit, objPlot):
			
		# If the unit that moved is a Great General, handle their move 
		# differently and return
		if(objUnit.getUnitType() == gc.getInfoTypeForString("UNIT_GENERAL")):
			self.handleGeneralMove(objUnit)
			return

		# Handle the movement for all non-Great General units.
		self.updatePlot(objUnit.getOwner(),objPlot)
		
		
	# Handles the moves made by a Great General unit
	def handleGeneralMove(self, objGeneral):
	
		# Return immediately if the unit passed in is invalid
		if(objGeneral == None):
			return
		
		# Return immediately if the unit passed in isn't really a general
		if(objGeneral.getUnitType() != gc.getInfoTypeForString("UNIT_GENERAL")):
			return
		
		# Save the Great Generals location
		self.saveGeneralLocation(objGeneral)

		# When the general moves update a 5x5 section of plots
		for dY in range(-3,4):		
			for dX in range(-3,4):
				# Don't process any plots that don't exist beyond the extents of the map.
				if(objGeneral.plot().getY()-dY >= 0 and objGeneral.plot().getY()+dY < gc.getMap().getGridHeight()):
				
					# Get the target plot
					objTargetPlot = gc.getMap().plot(objGeneral.plot().getX()+dX,objGeneral.plot().getY()+dY)

					# Update the plot
					self.updateGeneralPlot(objGeneral.getOwner(), objTargetPlot)


	# Removes all of the morale promotions given by the Great General passed in
	def removeAllGeneralPromotions(self, objGeneral):
	
		# Return immediately if the unit passed in isn't valid
		if(objGeneral == None):
			return
	
		# Return immediately if the unit passed in isn't a general
		if(objGeneral.getUnitType() != gc.getInfoTypeForString("UNIT_GENERAL")):
			return

		# If the Great General location information hasn't been previously 
		# logged, then set up the information.
		if(sdEntityExists("Great General Amped",objGeneral.getID()) == False):
			self.setupScript(objGeneral)

		iLastX = sdGetVal("Great General Amped", objGeneral.getID(), LAST_X)
		iLastY = sdGetVal("Great General Amped", objGeneral.getID(), LAST_Y)

		# Update a 5x5 section of plots with the general plot at the center.
		for dY in range(-3,4):		
			for dX in range(-3,4):
				# Don't process any plots that don't exist beyond the extents of the map.
				if(iLastY-dY >= 0 and iLastY+dY < gc.getMap().getGridHeight()):

					# Get the target plot
					objTargetPlot = gc.getMap().plot(iLastX+dX, iLastY+dY)
	
					# Update the target plot, removing all of the appropriate 
					# morale promotions and keeping the required ones, ok, so 
					# maybe it doesn't do exactly that, but it's close enough.
					self.updateGeneralPlot(objGeneral.getOwner(), objTargetPlot)
					