####################################
#
# TheLopezUtils.py
# Version: 0.1
# Copyright 2006
# This file provides a collection of useful methods for performing operations
# in Civilization 4.
#
# Version History:
# 
# v0.1:
# - Added the find global wonder method
# - Added the find national wonder method
# - Added the plot distance method
# - Added the unit distance method
# - Added the city distance method
# - Added the get plot unit list method
# - Added the get plot unit type list method
# - Added the get player plot unit list method
#
####################################

from CvPythonExtensions import *

import CvUtil
import PyHelpers

# globals
gc = CyGlobalContext()

PyPlayer = PyHelpers.PyPlayer
PyGame = PyHelpers.PyGame()

# Returns the city containing the nation wonder for the player ID passed in. 
# If a non-national unique build type ID is passed in the method will return 
# None.
def findNationalWonder(iBuildingTypeID, iPlayerID):
	'CyCity - city containing the global wonder'

	# Get the building information
	objBuildingInfo = gc.getBuildingInfo(iBuildingTypeID)

	# Get the player
	objPlayer = PyPlayer(iPlayerID)
	
	# Return None immediately if the building type ID passed in was invalid
	if(objBuildingInfo == None):
		return None
		
	# Get the building class information
	objBuildingClassInfo = gc.getBuildingClassInfo(objBuildingInfo.getBuildingClassType())
	
	# Return None immediately if the building isn't a unique global building
	if(objBuildingClassInfo.getMaxPlayerInstances() == -1):
		return None

	# Return None immediately if the player ID passed in is invalid.
	if(objPlayer == None):
		return None
				
	# Get the list of cities the player has
	cityList = objPlayer.getCityList()

	# Go through the players cities
	for objCity in cityList:

		# If the city has the global wonder then return the city
		if(objCity.hasBuilding(iBuildingTypeID)):
			return objCity.GetCy()

	return None
	

# Returns the city containing the global wonder. If a non-global unique build
# type ID is passed in the method will return None.
def findGlobalWonder(iBuildingTypeID):
	'CyCity - city containing the global wonder'

	# Get the building information
	objBuildingInfo = gc.getBuildingInfo(iBuildingTypeID)
	
	# Return None immediately if the building type ID passed in was invalid
	if(objBuildingInfo == None):
		return None
		
	# Get the building class information
	objBuildingClassInfo = gc.getBuildingClassInfo(objBuildingInfo.getBuildingClassType())
	
	# Return None immediately if the building isn't a unique global building
	if(objBuildingClassInfo.getMaxGlobalInstances() == -1):
		return None
			
	# Get all of the players in the game
	playerList = PyGame.getCivPlayerList()

	# Go through all the players in the game
	for objPlayer in playerList:

			# Get the list of cities the player has
			cityList = objPlayer.getCityList()

			# Go through the players cities
			for objCity in cityList:

				# If the city has the global wonder then return the city
				if(objCity.hasBuilding(iBuildingTypeID)):
					return objCity.GetCy()

	return None
	

# Returns the list of player cities containing at least one specialist of the 
# type passed in.
def getPlayerCitiesContainingSpecialist(iSpecialistTypeID, iPlayerID):
	'cityList - the list containing the cities with at least one specialist of the type passed in'

	# Get the player
	objPlayer = PyPlayer(iPlayerID)

	specialistCityList = []
	
	# Return None immediately if the player ID passed in is invalid.
	if(objPlayer == None):
		return None
				
	# Get the list of cities the player has
	cityList = objPlayer.getCityList()

	# Go through the players cities
	for objCity in cityList:

		# If the city has at least one specialists of the type passed in.
		if(objCity.GetCy().getFreeSpecialistCount(iSpecialistTypeID) > 0):
			specialistCityList.append(objCity.GetCy())

	return specialistCityList


# Calculates the distance between two plots using the formula:
# MAX(ABS(X2-X1), ABS(Y2-Y1))
def getPlotDistance(objPlotA, objPlotB):
	' iDist - the distance between objPlotA and objPlotB'
	dX = abs(objPlotB.getX()-objPlotA.getX())
	dY = abs(objPlotB.getY()-objPlotA.getY())
	dist = max(dX, dY)

	# if you want it in one line to avoid variable assignments: 
	# return max(abs(objPlotB.getX()-objPlotA.getX()), abs(objPlotB.getY()-objPlotA.getY()))
	return dist
	
	
# Calculates the distance between two units using the formula:
# MAX(ABS(X2-X1), ABS(Y2-Y1))
def getUnitDistance(objUnitA, objUnitB):
	' iDist - the distance between objUnitA and objUnitB'
	return getPlotDistance(objUnitA.plot(), objUnitB.plot())


# Calculates the distance between two cities using the formula:
# MAX(ABS(X2-X1), ABS(Y2-Y1))
def getCityDistance(objCityA, objCityB):
	' iDist - the distance between objCityA and objCityB'
	return getPlotDistance(objCityA.plot(), objCityB.plot())


# Returns the list of units in the plot passed in
def getPlotUnitList(objPlot):
	' unitList - All of the units in the objPlot'

	unitList = []

	# Get the number of units in the plot
	iUnitCount = objPlot.getNumUnits()

	# Go through the number of units in the plot
	for i in range(iUnitCount):
		# Get a unit from the plot
		objUnit = objPlot.getUnit(i)
			
		# If the unit isn't dead then add them to the list
		if(not objUnit.isDead()):
			unitList.append(objUnit)
	
	return unitList


# Returns the list of live units in a plot with their type set to iUnitType.
def getPlotUnitTypeList(objPlot, iUnitTypeID):
	' unitList - All of the units in the objPlot with the type set to iUnitType'
	unitList = []

	# Get the number of units in the plot
	iUnitCount = objPlot.getNumUnits()

	# Go through each unit found in the plot
	for i in range(iUnitCount):
		# Get a unit from the plot
		objUnit = objPlot.getUnit(i)
			
		# If the unit isn't dead and their unit type matches the 
		# iUnitType add them to the list
		if(not objUnit.isDead() and objUnit.getUnitType() == iUnitTypeID):
			unitList.append(objUnit)
				
				
# Returns the units in a plot for a particular player ID passed in
def getPlayerPlotUnitList(objPlot, iPlayerID):
	' unitList - All of the units in the objPlot'
	unitList = []

	# Get the number of units in the plot
	iUnitCount = objPlot.getNumUnits()
		
	for i in range(iUnitCount):
		# Get a unit from the plot
		objUnit = objPlot.getUnit(i)
	
		# If the unit isn't dead then add them to the list
		if(not objUnit.isDead() and objUnit.getOwner() == iPlayerID):
			unitList.append(objUnit)
		
	return unitList	
				
